-- Everything related to colour shifting
local espio = chaotix.espio

-- command so people can set their own shift colours!
COM_AddCommand(espio.ColourShiftCommand, function(p, ...)
	local shiftColours = {}
	local colours = {...}
	if not (#colours) 
		p.espio_shiftcolours = espio.ColourShiftDefault
		chaotix.WriteConfig(p)	-- update config
		return 
	end
	
	for k, v in ipairs(colours)
		local colour = R_GetColorByName(v)
		if not (colour) then continue end
		
		table.insert(shiftColours, colour)
	end
	if not (#shiftColours) then return end
	
	p.espio_shiftcolours = shiftColours
	chaotix.WriteConfig(p)	-- update config
end)

-- the function that handles it
local function espio_ghostCopy(ghost, mo)
	-- these need to be set directly
	ghost.target = mo
	ghost.tracer = mo
	ghost.fuse = 2
	ghost.scale = mo.scale
	ghost.destscale = mo.scale
	ghost.flags2 = (mo.flags2 & MF2_SPLAT) | MF2_LINKDRAW
	
	-- the rest can be copied pretty easily
	local fields = {
		"skin",
		"rollangle",
		"state", "tics",
		"sprite", "sprite2", "frame",
		"renderflags", "blendmode",
		"spritexscale", "spriteyscale", "spritexoffset", "spriteyoffset"
	}
	for _, field in ipairs(fields)
		ghost[field] = mo[field]
	end
	
	-- now move to our source object
	local health = mo.health
	mo.health = 1
	A_CapeChase(ghost)
	mo.health = health
end

espio.ColourShift = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	if (mo.skin != espio.SKIN) then mo.espio_shifttimer = nil return end
	
	-- conditions where we should *not* colour shift
	if ((gametyperules & GTR_TEAMS) and p.ctfteam)	-- no confusion in team gamemodes
	or (mo.espio_battleopacity != nil) -- no color shift while invisible
		mo.espio_shifttimer = nil
		return
	end
	
	-- erase the colour shift objects during super transformation
	if (mo.state == S_PLAY_SUPER_TRANS3 or mo.state == S_PLAY_NIGHTS_TRANS3)
	and (mo.espio_shifttimer != nil)
		mo.espio_shifttimer = nil
		for i = 1, 2
			local variable = "espio_shiftghost" + i
			local ghost = mo[variable]
			
			if not (ghost and ghost.valid) then continue end
			P_RemoveMobj(ghost)
		end
		return
	end
	
	-- conditions when we should start colour shifting
	if ((mo.state == S_PLAY_WAIT) or (mo.state == S_PLAY_NIGHTS_STAND)	-- we're bored or standing in nights
	and (mo.frame > 0))	-- and we're not in the transition frame
	//or (mo.espio_focuscharges == 5)	-- we're fully charged!
	//or (mo.espio_chargejump_timer and mo.espio_chargejump_timer > espio.ChargeJumpDuration)	-- we're fully charged!
	or (mo.espio_teleport_state == 2)	-- focus...
	or (mo.state == S_PLAY_EDGE)	-- we're on the edge
	or (Tether and mo.state == S_PLAY_TETHER_HOLD)	-- smart tether support!
		if (mo.espio_colourshift_delay)
			mo.espio_colourshift_delay = $-1
		else
			mo.espio_shifttimer = $ or 0
		end
	end
	
	if (mo.espio_shifttimer != nil)	-- if it's not nil, we can start
		-- get the colour table
		p.espio_shiftcolours = $ or espio.ColourShiftDefault
		
		-- handle the timer and get maxindex
		mo.espio_shifttimer = $ + 1
		local timer = mo.espio_shifttimer
		local shifttics = espio.ColourShiftTics
		local maxindex = #p.espio_shiftcolours + 1
		local fade = max(1, min(10, 10 - (((timer % shifttics) * 10) / shifttics)))
		
		-- get our objects and do the logic
		for i = 1, 2
			local variable = "espio_shiftghost" + i
			local ghost = mo[variable]
				
			-- create the object if need be
			if not (ghost and ghost.valid)
				ghost = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_GHOST) --chaotix.SpawnGhostMobj(mo)
				mo[variable] = ghost
			end
			espio_ghostCopy(ghost, mo)
			
			-- don't show if we're transparent, colorized, or invisible!
			if mo.frame & FF_TRANSMASK or mo.colorized
			or mo.flags2 & (MF2_DONTDRAW | MF2_SHADOW)
				ghost.flags2 = $ | MF2_DONTDRAW
				continue
			end
			
			-- the start colour, what we're fading FROM
			if i == 1
				local startindex = (timer / shifttics) % maxindex
				ghost.color = p.espio_shiftcolours[startindex] or mo.color
				
			-- the shift colour, what we're fading TO
			elseif (fade >= 1 and fade <= 9)
				local shiftindex = ((timer + shifttics) / shifttics) % maxindex	-- get shiftindex
				
				-- fade to create the shift effect
				ghost.frame = ($ & ~FF_TRANSMASK) | (fade << FF_TRANSSHIFT)
				ghost.color = p.espio_shiftcolours[shiftindex] or mo.color
				
			-- guess we're not showing up after all
			else
				ghost.flags2 = $ | MF2_DONTDRAW
			end
		end
		
		-- is the shift over?
		local maxshifttics = (shifttics * maxindex) - 1
		if (timer >= maxshifttics)	-- maybe the table length would change
			mo.espio_shifttimer = nil
			for i = 1, 2
				local variable = "espio_shiftghost" + i
				local ghost = mo[variable]
				
				if not (ghost and ghost.valid) then continue end
				P_RemoveMobj(ghost)
			end
		end
	end
end