rawset(_G, "chaotix", {})

// -----------
// definitions
// -----------

chaotix.skins = {}
chaotix.characterLibraries = {}
chaotix.debug = false
chaotix.dprint = function(string)
	if not (chaotix.debug) then return end
	print(string)
end

// ---------
// character system
// ---------

// add a character to the chaotix
function chaotix.AddCharacter(library)
	table.insert(chaotix.skins, library.SKIN)
	table.insert(chaotix.characterLibraries, library)
end

// if @player is playing as a chaotix character, returns their skin number, else returns -1
function chaotix.GetSkinIndex(player)
	for _, num in ipairs(chaotix.GetSkinIndexes())
		if player.skin == num
			return num
		end
	end
	return -1
end

// returns the table with all of the skin numbers of the chaotix characters
function chaotix.GetSkinIndexes()
	if not chaotix.skinIndexes
		chaotix.skinIndexes = {}
		for i, name in ipairs(chaotix.skinNames)
			chaotix.skinIndexes[i] = #skins[name]
		end
	end
	return chaotix.skinIndexes
end

// ---------
// freeslot system
// ---------

function chaotix.CheckSlot(item) // this function deliberately errors when a freeslot does not exist
	if _G[item] == nil // this will error by itself for states and objects
		error() // this forces an error for sprites, which do actually return nil for some reason
	end
end

function chaotix.SafeFreeslot(...)
	local returnitem
	local items = {...}
	for i, item in ipairs(items)
		if item == nil
			continue
		end
		
		if pcall(chaotix.CheckSlot, item)
			print("\131NOTICE:\128 " .. item .. " was not allocated, as it already exists.")
		elseif not returnitem
			returnitem = freeslot(item)
		else
			freeslot(item)
		end
	end
	
	return returnitem
end

// ---------
// hook system
// ---------

chaotix.hooks = {}

// runs all of the hooks associated with @hookType
chaotix.RunHook = function(key, hookType, ...)
	local returnValue
	for i, func in ipairs(chaotix.hooks[key][hookType])
		local ret = func(...)
		if ret ~= nil
			returnValue = ret
		end
	end
	return returnValue
end

// hooks @func so that it is called when @hookType is run
chaotix.AddHook = function(hookType, func, extraArg)
	local key = extraArg or MT_NULL
	local hookExists = false
	local hooks = chaotix.hooks[key]
	
	// if necessary, create hook table for this string/object type
	if not hooks
		hooks = {}
		chaotix.hooks[key] = hooks
	end
	
	// if necessary, create a new table in the hook table for this hook type, and hook it
	if not hooks[hookType]
		hooks[hookType] = {}
		addHook(hookType, function(...)
			return chaotix.RunHook(key, hookType, ...)
		end, extraArg)
	end
	
	// add this function to the hook's list
	table.insert(hooks[hookType], func)
end

// ---------
// general functions
// ---------

// makes @mobj crumble @fof
function chaotix.MobjBustFOF(mo, fof)
	local line = fof.master
	
	// bust the FOF
	EV_CrumbleChain(fof)
	
	// run a linedef executor?
	if line.flags & ML_EFFECT5
		P_LinedefExecute(P_AproxDistance(line.dx, line.dy) >> FRACBITS, mo, fof.target)
	end
end

// returns the closest point on a line, taking its vertices into account
function chaotix.ClosestPointOnLine(line, x, y)
	local v1 = line.v1
	local v2 = line.v2
	local minX = min(v1.x, v2.x)
	local maxX = max(v1.x, v2.x)
	local minY = min(v1.y, v2.y)
	local maxY = max(v1.y, v2.y)
	
	x, y = P_ClosestPointOnLine(x, y, line)
	
	x = min(max($, minX), maxX)
	y = min(max($, minY), maxY)
	
	return x, y
end

// busts any breakable blocks within @mo's height range in the sector on the other side of @line
function chaotix.MobjBustFOFsThroughLine(mo, line, weak)
	local sector
	
	if P_PointOnLineSide(mo.x - mo.momx, mo.y - mo.momy, line) == 1
		sector = line.frontsector
	else
		sector = line.backsector
	end
	
	if sector == nil
		return
	end
	
	chaotix.MobjBustFOFsInSector(mo, mo.z, mo.z + mo.height, sector, weak)
end

// has @mo bust any breakable blocks between @bottom and @top in @sector
function chaotix.MobjBustFOFsInSector(mo, bottom, top, sector, weak)
	for fof in sector.ffloors()
		if fof.flags & (FF_EXISTS|FF_BUSTUP) ~= (FF_EXISTS|FF_BUSTUP)
		or weak and fof.flags & FF_STRONGBUST
			continue
		end
		
		local fofTop, fofBottom = chaotix.GetFOFHeightsAt(fof, mo.x, mo.y)
		
		if bottom < fofTop
		and top > fofBottom
			chaotix.MobjBustFOF(mo, fof)
		end
	end
end

// returns the top and bottom heights of @fof at the coordinates @x and @y
function chaotix.GetFOFHeightsAt(fof, x, y)
	local top
	local bottom
	
	if fof.t_slope
		top = P_GetZAt(fof.t_slope, x, y)
	else
		top = fof.topheight
	end
	
	if fof.b_slope
		bottom = P_GetZAt(fof.b_slope, x, y)
	else
		bottom = fof.bottomheight
	end
	
	return top, bottom
end

// checks for bustable FOFs above or below @player and busts them when close enough
function chaotix.CheckVerticalBustables(player)
	local mo = player.mo
	local x, y = mo.x, mo.y
	
	for fof in mo.subsector.sector.ffloors()
		if fof.flags & (FF_EXISTS|FF_BUSTUP) ~= (FF_EXISTS|FF_BUSTUP)
			continue
		end
		
		local topZ, bottomZ = chaotix.GetFOFHeightsAt(fof, x, y)
		
		if mo.z + mo.momz <= topZ
		and mo.z + mo.height + mo.momz >= bottomZ
			chaotix.MobjBustFOF(mo, fof)
		end
	end
end

-- helper function for setting frames (accounts for frameflags and other extra stuff)
chaotix.SetFrame = function(mo, frame, sprite, noflags)
	local frameflags = not (noflags) and (mo.frame & ~FF_FRAMEMASK) or 0
	-- ^ if noflags is checked, dont keep frameflags when setting the frame
	
	if (sprite)
		local spr = (mo.player) and "sprite2" or "sprite"	-- players use sprite2
		mo[spr] = sprite
	end
	mo.frame = frame|frameflags
end

-- I am SICK of doing this so many times, laziness time
chaotix.WaterMom = function(mo, momz, modifier)
	modifier = $ or FRACUNIT/2
	return FixedMul(momz, (mo.eflags & MFE_UNDERWATER) and modifier or FRACUNIT)
end

-- angtoint
chaotix.AngToInt = function(a)
	if not a return 0 end
	a = $/ANG1
	if a < 0
		a = $+360
	end
	return a
end

-- function to check lines
chaotix.CheckLine = function(p, line, allownoclimb)
	local mo = p.mo
	
	-- really easy one, dont cling to noclimb walls
	if not (allownoclimb)
		if (line.flags & ML_NOCLIMB) then return false end
	end
	
	-- horizon line
	if (line.special == 41) then return end
	
	-- sector checks
	if not (line.backsector) then return true end	-- apparently in these situations source always returns true?
	local sec = (line.backsector and not P_PointOnLineSide(mo.x, mo.y, line)) and line.backsector or line.frontsector
	chaotix.dprint((sec == line.backsector) and "backsector" or "frontsector")
	
	local canclimb = false
	local checkz = {mo.z, mo.z + mo.height}
	if (mo.eflags & MFE_VERTICALFLIP) then checkz = {mo.z + mo.height, mo.z} end
	
	-- floor check
	chaotix.dprint("Floor CheckZ: "..checkz[2])
	chaotix.dprint("Floor Height: "..sec.floorheight/FRACUNIT)
	if (checkz[2] < sec.floorheight)
		canclimb = true 
		chaotix.dprint("floor check") 
	end
	
	-- ceiling checks
	chaotix.dprint("Ceiling CheckZ: "..checkz[1])
	chaotix.dprint("Ceiling Height: "..sec.ceilingheight/FRACUNIT)
	local thokbarrier = (sec.ceilingpic:upper():find("F_SKY"))	-- walls above this ceiling arent rendered, shouldnt be able to climb?
	if ((checkz[1] >= sec.ceilingheight) and not thokbarrier) 
		canclimb = true
		chaotix.dprint("ceiling check") 
	end
	
	-- FOF checks
	for rover in sec.ffloors()
		if not (rover.flags & FF_EXISTS or rover.flags & FF_BLOCKPLAYER) then continue end
		
		if (checkz[1] > rover.bottomheight)
		and (checkz[2] < rover.topheight)
			canclimb = true
			chaotix.dprint("FOF check")
			break
		end
	end
	
	-- if we're here, line checks went right
	return canclimb
end

-- thanks to interp, gotta spawn ghosts myself...
chaotix.SpawnGhostMobj = function(mo)
	local ghost = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_GHOST)
	
	-- these need to be set directly
	ghost.scale = mo.scale
	ghost.destscale = mo.scale
	ghost.angle = mo.player and mo.player.drawangle or mo.angle
	
	-- the rest can be copied pretty easily
	local fields = {
		"skin",
		"rollangle",
		"sprite", "sprite2", "frame",
		"renderflags", "blendmode",
		"spritexscale", "spriteyscale", "spritexoffset", "spriteyoffset"
	}
	for _, field in ipairs(fields)
		ghost[field] = mo[field]
	end
	
	return ghost
end

chaotix.RandomElement = function(t)
	return t[P_RandomRange(1,#t)]
end

chaotix.FixedLerp = function(val1, val2, amt)
	return FixedMul(FRACUNIT-amt,val1) + FixedMul(amt,val2)
end

chaotix.PlayerSpeedLimit = function(player, topspeed, friction)
	local spd = R_PointToDist2(0,0,player.rmomx,player.rmomy)
	if spd > topspeed
		local factor = FixedDiv(max(topspeed, FixedMul(spd,friction)), spd)
		player.rmomx = FixedMul(player.rmomx, factor)
		player.rmomy = FixedMul(player.rmomy, factor)
		player.mo.momx = player.rmomx + player.cmomx
		player.mo.momy = player.rmomy + player.cmomy
		return true
	end
	return false
end

//Return the angle direction produced by player move commands
chaotix.GetInputAngle = function(player)
	local fw = player.cmd.forwardmove
	local sw = player.cmd.sidemove
-- 	local pang = player.cmd.angleturn << 16//is this netsafe?
	local pang = player.mo.angle
	
	if fw == 0 and sw == 0 then
		return pang
	end
	
	local c0, s0 = cos(pang), sin(pang)
	
	local rx, ry = fw*c0 + sw*s0, fw*s0 - sw*c0
	local retangle = R_PointToAngle2(0, 0, rx, ry)
	return retangle
end

chaotix.GetInputThrust = function(player)
	local fw = player.cmd.forwardmove*FRACUNIT/50
	local sw = player.cmd.sidemove*FRACUNIT/50
	return R_PointToDist2(0, 0, fw, sw)
end

-- removes an empty caption from sfxinfo
chaotix.RemoveEmptyCaption = function(info)
	if (info.caption != "") then return end
	info.caption = "/"
end

chaotix.SpawnSpeedTrail = function(mo, condition, v)
	if condition == 0 or condition == false then return end -- set condition
	if v == nil then v = {} end
	
	-- get momentum
	if v.momx == nil then v.momx = mo.momx end
	if v.momy == nil then v.momy = mo.momy end
	if v.momz == nil then v.momz = mo.momz end
	
	-- set up the random positioning
	local hdist = (v.radius or mo.radius*2)/mo.scale
	local vdist = (v.height or mo.height)/mo.scale
	vdist = P_RandomRange(0, $)*FRACUNIT
	
	-- get the angle
	local angle = mo.angle
	if v.momx or v.momy
		angle = R_PointToAngle2(0,0, v.momx, v.momy)
	elseif mo.player
		angle = mo.player.drawangle
	end
	
	-- spawn the trail
	local thok = P_SpawnMobjFromMobj(mo,
	P_RandomRange(-hdist, hdist)*FRACUNIT,
	P_RandomRange(-hdist, hdist)*FRACUNIT,
	vdist, MT_THOK)
	
	-- set the angles and state
	thok.angle = angle
	thok.rollangle = R_PointToAngle2(0,0, FixedHypot(v.momx, v.momy), v.momz)
	thok.state = S_CHAOTIX_SPEEDTRAIL
	
	-- set scale and color
	if v.scale then thok.scale = v.scale end
	if v.color
		thok.color = v.color
		thok.colorized = true
	end
	
	-- thrust back a little
	local factor = -4*thok.scale
	P_Thrust(thok, angle, FixedMul(abs(cos(thok.rollangle)), factor))
	P_SetObjectMomZ(thok, FixedMul(sin(thok.rollangle), factor))
end

/*
chaotix.AddHook("PlayerThink", function(p)
	if p.cmd.buttons & BT_CUSTOM1 and p.mo
		chaotix.SpawnSpeedTrail(p.mo, leveltime & 1)
	end
end)
*/